/* dnsc - dns client module */

#include "Python.h"

#include "dns.h"
#include "ip4.h"

#define VERSION         "0.11"
#define DJBDNSVERSION   "1.05"

char * str_z;

static stralloc in   = {0};	/* so named after examples in djb's docs */
static stralloc out  = {0};	/* see <http://cr.yp.to/djbdns/dns.html>  */
static stralloc fqdn = {0};

PyObject * DnscError;

#define DNSC_FUNCTION_HEADER(funcname)			\
static PyObject *					\
dnsc_##funcname##(PyObject *ignore, PyObject *args) {	\



#define DNSC_PARSEARG(funcstr)				\
							\
    if (! PyArg_ParseTuple(args, "s:"funcstr, &str_z))	\
        return NULL;					\
							\
    if (! stralloc_copys(&in, str_z))			\
        return PyErr_SetFromErrno(DnscError);		\


/* ipsa2tuple(ip_sa) -> ( '1.2.3.4', ... ) */

static PyObject *
ipsa2tuple(stralloc *ip) {
    PyObject * tuple;
    char ascii_dotted[16];
    int i, j;

    tuple = PyTuple_New(ip->len / 4); /* trust ip->len % 4 == 0 */

    if (ip->len) {
        for (i = 0;i*4 < ip->len; i++) {
            j = ip4_fmt(ascii_dotted, (char *)(ip->s) + (i*4));
            PyTuple_SetItem(tuple, i,
                           Py_BuildValue("s#", ascii_dotted, j));
        }
    }
    return tuple;
}

static char dnsc_ip4__doc__[] = "\
ip4(fqdn) -> (ipv4, ...)\n\
\n\
Return a tuple of IPv4 address strings found in DNS for the fully-\n\
qualified domain name fqdn.";

DNSC_FUNCTION_HEADER(ip4)

    DNSC_PARSEARG("ip4")

    if (dns_ip4(&out, &in) == -1)
        return PyErr_SetFromErrno(DnscError);

    return ipsa2tuple(&out);

}

static char dnsc_ns__doc__[] = "\
ns(fqdn) -> (ns, ...)\n\
Return the NS records for a given fqdn.";

DNSC_FUNCTION_HEADER(ns)
    PyObject * tuple;

    DNSC_PARSEARG("ns")

    if (dns_ns(&out, &in) == -1)
        return PyErr_SetFromErrno(DnscError);

    if (out.len) {
        char * s;
        int i = 0;
        for (s = out.s; s - out.s < out.len; i++) {
            s += strlen(s) + 1;
        }
        tuple = PyTuple_New(i);
        if (tuple == NULL)
            return NULL;
        i = 0;
        for (s = out.s; s - out.s < out.len; i++) {
            PyTuple_SetItem(tuple, i, Py_BuildValue("s", s));
            s += strlen(s) + 1;
        }
    } else {
        tuple = PyTuple_New(0);
        if (tuple == NULL)
            return NULL;
    }

    return tuple;
}

static char dnsc_mx__doc__[] = "\
mx(fqdn) -> ((distance, dn), ...)\n\
\n\
Look up mail exchanger records for the fully-qualified domain name\n\
fqdn, returning a tuple of (distance, mx_record) pairs.";

DNSC_FUNCTION_HEADER(mx)
    PyObject * tuple, *i;
    char * s;
    int preference, mx_count = 0;

    DNSC_PARSEARG("mx")

    if (dns_mx(&out, &in) == -1)
        return PyErr_SetFromErrno(DnscError);

    if (out.len) {
        for (s = out.s; s - out.s < out.len; mx_count++) {
            s += 2;
            s += strlen(s) + 1;
        }

        tuple = PyTuple_New(mx_count);
        if (tuple == NULL)
            return NULL;

        mx_count = 0;

        for (s = out.s; s - out.s < out.len; mx_count++) {
            preference = 0 + ((unsigned char)s[0]<<8) + (unsigned char)s[1];
            s += 2;
            PyTuple_SetItem(tuple, mx_count,
                Py_BuildValue("(is)", preference, s));
            s += strlen(s) + 1;
        }
    } else {  /* out.len == 0; no MX records found */
        tuple = Py_BuildValue("()");
    }

    return tuple;
}

static char dnsc_txt__doc__[] = "\
txt(fqdn) -> string (or None)\n\
\n\
Look up TXT records for fully-qualified domain name fqdn,\n\
returning their string concatentation.";

DNSC_FUNCTION_HEADER(txt)

    DNSC_PARSEARG("txt")

    if (dns_txt(&out, &in) == -1)
        return PyErr_SetFromErrno(DnscError);

    return Py_BuildValue("s#", out.s, out.len);
}

static char dnsc_ip4_qualify__doc__[] = "\
ip4_qualify(udn) -> (fqdn, (ipv4, ...))\n\
\n\
Lookup addresses for qualifications of the name udn, returning\n\
the first fully qualified match in the DNS and a tuple of its\n\
dotted IPv4 addresses.\n\
\n\
Qualification supports traditional hostname rewriting schemes:\n\
os.environ['LOCALDOMAIN'], /etc/resolv.conf (reread every 10\n\
minutes for changes), and gethostname().\n\
\n\
A powerful, user-controlled qualification scheme is also offered\n\
via /etc/dnsrewrite (or os.environ['DNSREWRITEFILE']).  See\n\
http://cr.yp.to/djbdns/qualify.html for details.";

DNSC_FUNCTION_HEADER(ip4_qualify)
    PyObject * tuple;

    DNSC_PARSEARG("ip4_qualify")

    if (dns_ip4_qualify(&out, &fqdn, &in) == -1)
        return PyErr_SetFromErrno(DnscError);

    tuple = PyTuple_New(2);

    PyTuple_SetItem(tuple, 0,
                   Py_BuildValue("s#", fqdn.s, fqdn.len));

    PyTuple_SetItem(tuple, 1, ipsa2tuple(&out));

    return tuple;
}

static char dnsc_name4__doc__[] = "\
name4(ipv4) -> dns_name (or None)\n\
\n\
Look up the domain name for the dotted address string ipv4.";

DNSC_FUNCTION_HEADER(name4)

    char octets[4];

    if (! PyArg_ParseTuple(args, "s:name4", &str_z))
        return NULL;

    if (! ip4_scan(str_z, octets)) {
        (void) PyErr_SetString(
                   PyExc_TypeError,
                   "Function expects ascii dotted IPv4 address");
        return NULL;
    }

    if (dns_name4(&out, octets) == -1)
        return PyErr_SetFromErrno(DnscError);

    return out.len ?
           Py_BuildValue("s#", out.s, out.len) :
           Py_BuildValue("");
}

static char dnsc_ip4_pack__doc__[] = "\
ip4_pack(ipv4) -> packed 32-bit IP representation\n\
\n\
Convert a dotted IPv4 string to its 32-bit packed representation\n\
(e.g., ip4_pack('127.0.0.1') -> '\\177\\000\\000\\001').";

DNSC_FUNCTION_HEADER(ip4_pack)

    char octets[4];

    if (! PyArg_ParseTuple(args, "s:ip4_pack", &str_z))
        return NULL;

    if (! ip4_scan(str_z, octets)) {
        (void) PyErr_SetString(
                   PyExc_TypeError,
                   "Function expects dotted IPv4 address string");
        return NULL;
    }

    return PyString_FromStringAndSize(octets, 4);

}

static char dnsc_ip4_unpack__doc__[] = "\
ip4_unpack(packed_addr) -> dotted IPv4 string\n\
\n\
Convert an IPv4 address from its 32-bit packed binary form to\n\
dotted string notation (e.g., ip4_unpack('\\177\\000\\000\\001') ->\n\
'127.0.0.1').";

DNSC_FUNCTION_HEADER(ip4_unpack)

    char * dotted;
    int i;

    if (! PyArg_ParseTuple(args, "s#:ip4_unpack", &str_z, &i))
        return NULL;

    if (i != 4) {
        (void) PyErr_SetString(
                   PyExc_TypeError,
                   "Function expects octet string of 4 bytes");
        return NULL;
    }

    i = ip4_fmt(dotted, str_z);

    return PyString_FromStringAndSize(dotted, i);

}

static PyMethodDef DnscMethods[] = {

    {"ip4",          dnsc_ip4,
                     METH_VARARGS,
                     dnsc_ip4__doc__},

    {"mx",           dnsc_mx,
                     METH_VARARGS,
                     dnsc_mx__doc__},

    {"ns",           dnsc_ns,
                     METH_VARARGS,
                     dnsc_ns__doc__},

    {"txt",          dnsc_txt,
                     METH_VARARGS,
                     dnsc_txt__doc__},

    {"ip4_qualify",  dnsc_ip4_qualify,
                     METH_VARARGS,
                     dnsc_ip4_qualify__doc__},

    {"name4",        dnsc_name4,
                     METH_VARARGS,
                     dnsc_name4__doc__},

    {"ip4_pack",     dnsc_ip4_pack,
                     METH_VARARGS,
                     dnsc_ip4_pack__doc__},

    {"ip4_unpack",   dnsc_ip4_unpack,
                     METH_VARARGS,
                     dnsc_ip4_unpack__doc__},

    {NULL, NULL}
};

DL_EXPORT(void)
initdnslook() {
  PyObject *m, *d;

  m = Py_InitModule3("dnslook", DnscMethods,
"This module offers several easy-to-use, high-level DNS lookup routines\n\
adapted from D. J. Bernstein's djbdns package.\n\
\n\
One exception, \"error\", is defined, with a value corresponding to the\n\
values of OSError/os.error exceptions.  As in the underlying library,\n\
drastic query failure (e.g., malformed response to the final query\n\
attempt) matches the value of errno.EIO, and server failure matches that\n\
of errno.EAGAIN.\n\
\n\
The routines are all stub resolvers, sending recursive DNS queries to a\n\
nearby DNS cache.  /etc/resolv.conf's 'nameserver' line (overridden by\n\
os.environ['DNSCACHEIP'], defaulting to 127.0.0.1) sets the cache's IP.\n\
\n\
See http://cr.yp.to/djbdns/dns.html for details on the underlying\n\
library.");

  d = PyModule_GetDict(m);

  DnscError = PyErr_NewException("dnslook.error", NULL, NULL);
  PyDict_SetItemString(d, "error", DnscError);

  PyDict_SetItemString(d, "__version__", 
                       PyString_FromString(VERSION));

  PyDict_SetItemString(d, "__djbdns_version__",
                       PyString_FromString(DJBDNSVERSION));

}
